package at.ac.uibk.gitsearch.repository.search;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeMap;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PostConstruct;
import javax.ws.rs.NotFoundException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.MatchQueryBuilder;
import org.elasticsearch.index.query.MultiMatchQueryBuilder;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightBuilder;
import org.springframework.context.event.ContextStoppedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Repository;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import at.ac.uibk.gitsearch.config.ApplicationProperties;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO;
import at.ac.uibk.gitsearch.service.dto.UserProvidedMetadataDTO.Person;

@Repository
public class MetaDataRepository {

	private final RestHighLevelClient elasticsearchClient;

	private static final Logger LOGGER = LogManager.getLogger(MetaDataRepository.class);

	private final ApplicationProperties properties;

	private static final int FRAGMENT_SIZE = 5000;

	public MetaDataRepository(RestHighLevelClient elasticsearchClient, ApplicationProperties properties) {
		this.elasticsearchClient = elasticsearchClient;
		this.properties = properties;

	}

	/**
	 * contains for each Field, a list of Strings mapped to a completion. Together
	 * with hit counts E.g. "metadata.creator" -> Kerstin -> {"Kerstin Limbeck" ->
	 * 3, "Kerstin Mair" -> 2)}
	 */
	private Map<String, Map<String, Map<String, Integer>>> cachedCompletions = null;

	private static final int MAX_AUTO_COMPLETION_RESULTS = 10;

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix) throws IOException {
		return getFieldAutoCompletion(keyWordPrefix, SearchRepositoryConstants.METADATA_KEYWORDS);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param progammingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String progammingLanguagePrefix) throws IOException {
		return getFieldAutoCompletion(progammingLanguagePrefix,
				SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix) throws IOException {
		return getFieldAutoCompletion(creatorPrefix, SearchRepositoryConstants.METADATA_CREATOR);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix) throws IOException {
		return getFieldAutoCompletion(contributorPrefix, SearchRepositoryConstants.METADATA_CONTRIBUTOR);
	}

	/**
	 * returns all contributor and creator autocompletes
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix) throws IOException {
		final List<AutoCompleteEntry> contributors = getContributorAutoComplete(contributorPrefix);
		final List<AutoCompleteEntry> creatorAutoComplete = getCreatorAutoComplete(contributorPrefix);
		// merging
		for(AutoCompleteEntry creator: creatorAutoComplete) {
			contributors.stream().filter(entry -> entry.getTarget().equals(creator.getTarget())).findAny()
			   .ifPresentOrElse(entry -> {contributors.remove(entry); 
			                             contributors.add(new AutoCompleteEntry(creator.getTarget(), creator.getHitCount() + entry.getHitCount()));}, 
					   () -> contributors.add(creator));
		}
		return contributors;
	}
	/**
	 * returns a list of hits together with the number of occurences
	 * @param keyWordPrefix
	 * @param metaDataField
	 * @return
	 * @throws IOException
	 * @throws JsonProcessingException
	 * @throws JsonMappingException
	 */
	private List<AutoCompleteEntry> getFieldAutoCompletion(String keyWordPrefix, final String metaDataField)
			throws IOException, JsonProcessingException, JsonMappingException {
		fillAutoCompletion();

		String lcKeyWordPrefix = keyWordPrefix.toLowerCase();
		// höllisch kompliziert und höllisch unverständlich mit Streams :-)
		// idea
		// 1. Filtern nach Prefix
		// 2. Zusammenzählen der möglichen Hits (Besser Max statt Sum?)
		// 3. Auf MAX_AUTO_COMPLETION_RESULTS begrenzen und nur die Trefferstrings
		// ausgeben.
		final Stream<Entry<String, Map<String, Integer>>> filteredEntries = cachedCompletions.get(metaDataField)
				.entrySet().stream() // Map<String, Map<String, Integer>> -> Stream<Entry<String, Map<String,
										// Integer>>
				.filter(s -> s.getKey().startsWith(lcKeyWordPrefix));
		final Map<String, Integer> combinedHits = filteredEntries // s = String -> bool -> Stream<Entry<String,
																	// Map<String, Integer>>
				.flatMap(e -> e.getValue().entrySet().stream()) // Stream< Map<String, Integer>> with duplicates
				.collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue, Integer::max));
		return combinedHits.entrySet().stream().sorted((e1, e2) -> e2.getValue() - e1.getValue())
				.limit(MAX_AUTO_COMPLETION_RESULTS).map(s -> new AutoCompleteEntry(s.getKey(), s.getValue())).collect(Collectors.toList()); // Map.Entry<String,
																										// Map<String,
																										// Integer>>

	}

	private static final long AUTOCOMPLETION_RECHECK_TIMEOUT = 5 * 60 * 1000L; // every 5 Minutes
	private Timer autocompletionReloadTimer = new Timer("AutocompletionReloadTimer");
	@PostConstruct
	private void setUpAutocompletion() {
		
		autocompletionReloadTimer.schedule(new TimerTask() {

			@Override
			public void run() {
				try {
					updateAutocompletionCache();
				} catch(IOException e) {
					LOGGER.warn("Cannot refresh Autocompletion Cache", e);
				}
			}
		}, 0L, AUTOCOMPLETION_RECHECK_TIMEOUT);
	}
	
	/**
	 * stops the cache reload timer, during shutdown or between tests.
	 * @param event ContextStoppedEvent
	 */
    @EventListener
    protected void stopReloadTimer(ContextStoppedEvent event) {
    	autocompletionReloadTimer.cancel();
    }


	private synchronized void fillAutoCompletion() throws IOException, JsonProcessingException, JsonMappingException {
		if (cachedCompletions == null) {
			updateAutocompletionCache();
		}
	}

	private void updateAutocompletionCache() throws IOException, JsonProcessingException, JsonMappingException {
		Map<String, Map<String, Map<String, Integer>>> reloadedCachedCompletions = new HashMap<>();
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_KEYWORDS, new TreeMap<>());
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CREATOR, new TreeMap<>());
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_CONTRIBUTOR, new TreeMap<>());
		reloadedCachedCompletions.put(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, new TreeMap<>());

		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		final int pageSize = 20;
		int currentPage = 0;
		boolean tryNextPage = true;

		while (tryNextPage) {

			SearchSourceBuilder sourceBuilder = new SearchSourceBuilder().size(pageSize).from(currentPage++*pageSize);

			searchRequest.source(sourceBuilder);

			SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

			ObjectMapper objectMapper = new ObjectMapper();
			objectMapper.registerModule(new JavaTimeModule());
			objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

			tryNextPage = searchResponse.getHits().getHits().length > 0;

			for (SearchHit searchHit : searchResponse.getHits().getHits()) {
				final String sourceAsJSON = searchHit.getSourceAsString();
				SearchResultDTO entry = objectMapper.readValue(sourceAsJSON, SearchResultDTO.class);
				if (entry.getMetadata().getKeyword() != null)
					for (String keyWord : entry.getMetadata().getKeyword())
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_KEYWORDS),
								split(keyWord), keyWord);
				if (entry.getMetadata().getContributor() != null)
					for (Person contributor : entry.getMetadata().getContributor())
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CONTRIBUTOR),
								split(contributor.getName()), contributor.getName());
				if (entry.getMetadata().getCreator() != null)
					for (Person creator : entry.getMetadata().getCreator())
						addTo(reloadedCachedCompletions.get(SearchRepositoryConstants.METADATA_CREATOR),
								split(creator.getName()), creator.getName());
				if (entry.getMetadata().getProgrammingLanguage() != null)
					for (String language : entry.getMetadata().getProgrammingLanguage())
						addTo(reloadedCachedCompletions
								.get(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES),
								Collections.singletonList(language), language);
			}
		}
		cachedCompletions = reloadedCachedCompletions;
	}

	private static void addTo(Map<String, Map<String, Integer>> completionMap, List<String> tokenList, String target) {
		tokenList.forEach(token -> {
			String lowercaseToken = token.toLowerCase();
			Map<String, Integer> existingTargets = completionMap.get(lowercaseToken);
			if (existingTargets == null) { // none found
				existingTargets = new HashMap<>();
				existingTargets.put(target, 1);
				completionMap.put(lowercaseToken, existingTargets);
			} else {
				final Integer count = existingTargets.get(target);
				if (count == null) {
					existingTargets.put(target, 1);
				} else {
					existingTargets.put(target, count + 1);
				}

			}
		});
	}

	private static List<String> split(String s) {
		StringTokenizer st = new StringTokenizer(s, " \t\n\r\f");
		List<String> result = new ArrayList<>();
		while (st.hasMoreTokens()) {
			result.add(st.nextToken());
		}
		return result;
	}

	/**
	 * returns the hits from the meta data search index for the input query parameters.
	 * @param searchInputDTO the query parameters.
	 * @param pageSize the size of the page (i.e. the maximum of returned hits).
	 * @param user the user that queries.
	 * @return the search hits.
	 * @throws IOException when search index is not available.
	 */
	public SearchResultsDTO pageDetails(SearchInputDTO searchInputDTO, int pageSize, Optional<User> user) throws IOException {
		fillAutoCompletion();
		
		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();

		BoolQueryBuilder fullTextBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getFulltextQuery(),
				term -> fullTextBuilder.should(QueryBuilders
						.multiMatchQuery(term, SearchRepositoryConstants.METADATA_DESCRIPTION,
								SearchRepositoryConstants.METADATA_KEYWORDS, SearchRepositoryConstants.METADATA_TITLE)
						.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX)));
		if(fullTextBuilder.hasClauses())
			queryBuilder.must(fullTextBuilder);

		BoolQueryBuilder plBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getMetadata().getProgrammingLanguage(), term -> plBuilder
				.should(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, term)));
		if(plBuilder.hasClauses())
			queryBuilder.must(plBuilder);

		BoolQueryBuilder keywordBuilder = QueryBuilders.boolQuery();
		forEachTerm(searchInputDTO.getMetadata().getKeyword(), term -> keywordBuilder
				.should(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_KEYWORDS, term)));
		if(keywordBuilder.hasClauses())
			queryBuilder.must(keywordBuilder);

		BoolQueryBuilder authorBuilder = QueryBuilders.boolQuery();
		if (searchInputDTO.getMetadata().getAuthor() != null)
			authorBuilder.should(QueryBuilders
					.multiMatchQuery(searchInputDTO.getMetadata().getAuthor(),
							SearchRepositoryConstants.METADATA_CREATOR, SearchRepositoryConstants.METADATA_CONTRIBUTOR,
							SearchRepositoryConstants.METADATA_PUBLISHER)
					.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX));
		if(authorBuilder.hasClauses())
			queryBuilder.must(authorBuilder);
		
		if (searchInputDTO.getMetadata().getTypes()!=null && !searchInputDTO.getMetadata().getTypes().isEmpty()) {
			BoolQueryBuilder typeBuilder = QueryBuilders.boolQuery();
			searchInputDTO.getMetadata().getTypes().forEach(
						type -> typeBuilder.should(QueryBuilders.matchQuery(SearchRepositoryConstants.METADATA_TYPE, type.getExternalName())));
			if(typeBuilder.hasClauses())
				queryBuilder.must(typeBuilder);
		}

		forEachTerm(searchInputDTO.getMetadata().getLicense(),
				term -> queryBuilder.must(QueryBuilders.prefixQuery(SearchRepositoryConstants.METADATA_LICENSE, term)));

		// Authorization restrictions
		final MatchQueryBuilder publicQuery = QueryBuilders.matchQuery(SearchRepositoryConstants.PROJECT_VISIBILITY, "public")
				.fuzzyTranspositions(false).autoGenerateSynonymsPhraseQuery(false).boost(0.0f);
		if(user.isEmpty()) {
			 queryBuilder
				.must(publicQuery);
		} else {
			
			final Collection<GrantedAuthority> authorities = user.get().getAuthorities();
			final BoolQueryBuilder authQuery = QueryBuilders.boolQuery().boost(0.0f);
			final Stream<QueryBuilder> prefixAuthQueries = authorities.stream()
			  .filter(auth -> !auth.getAuthority().startsWith("ROLE"))
			  .filter(auth -> !auth.getAuthority().startsWith("SCOPE"))
			  .map(auth -> QueryBuilders.prefixQuery(SearchRepositoryConstants.PROJECT_NAMESPACE, auth.getAuthority()));
			prefixAuthQueries.forEach(pq -> authQuery.should(pq) );
			if(authQuery.hasClauses())
				queryBuilder.must(QueryBuilders.boolQuery().should(authQuery).should(publicQuery));
			else
				queryBuilder.must(publicQuery);
		}

		char[] boundaryChars = { '\n' };

		HighlightBuilder highlightBuilder = new HighlightBuilder().highlighterType("plain")
				.preTags(properties.getSearch().getHighlightPre()).postTags(properties.getSearch().getHighlightPost())
				.field(SearchRepositoryConstants.METADATA_DESCRIPTION).boundaryChars(boundaryChars)
				.boundaryScannerType(HighlightBuilder.BoundaryScannerType.SENTENCE).fragmentSize(FRAGMENT_SIZE);

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder).highlighter(highlightBuilder).size(pageSize)
				.from((searchInputDTO.getPage()) * pageSize);

		searchRequest.source(sourceBuilder);

		SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

		return parseSearchResponse(searchResponse);
	}

	private static SearchResultsDTO parseSearchResponse(final SearchResponse searchResponse) throws JsonProcessingException {
		
		float maxScore = searchResponse.getHits().getMaxScore();
        ObjectMapper objectMapper = new ObjectMapper();
        objectMapper.registerModule(new JavaTimeModule());
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        final List<SearchResultDTO> searchResults = new ArrayList<>();
        for (SearchHit searchHit : searchResponse.getHits().getHits()) {
			float hitScore = searchHit.getScore();
            final SearchResultDTO parsedSearchHit = parseSearchHit(searchHit, objectMapper);
        	int ranking5 = 1;
        	if (maxScore>0.0f) ranking5 = (int) ((hitScore*5.0)/maxScore+0.5f);
			parsedSearchHit.setRanking5(ranking5);
			searchResults.add(parsedSearchHit);
        }

        long hitCount = searchResponse.getHits().getTotalHits();
        final SearchResultsDTO results = new SearchResultsDTO();
        results.setHitCount(hitCount);
        results.setSearchResult(searchResults);
        return results;
    }

    private static SearchResultDTO parseSearchHit(SearchHit searchHit, ObjectMapper objectMapper) throws JsonProcessingException {
        final SearchResultDTO mappedValue = objectMapper.readValue(searchHit.getSourceAsString(), SearchResultDTO.class);
        mappedValue.setExerciseId(searchHit.getId());
		return mappedValue;
    }

	/** Helper **/
	private static void forEachTerm(String searchTerm, Consumer<String> exec) {
		if (searchTerm == null)
			return;
		StringTokenizer st = new StringTokenizer(searchTerm);
		while (st.hasMoreTokens()) {
			exec.accept(st.nextToken());
		}
	}

	public SearchResultDTO getExerciseById(String exerciseId) throws NotFoundException {
		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();

		queryBuilder.must(QueryBuilders.matchQuery(SearchRepositoryConstants.EXERCISE_ID, exerciseId));

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder).size(10);

		searchRequest.source(sourceBuilder);

		try {
			SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

			ObjectMapper objectMapper = new ObjectMapper();
			objectMapper.registerModule(new JavaTimeModule());
			objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

			if (searchResponse.getHits().totalHits > 1) {
				LOGGER.warn("Found {} hits for exercise id {}, expected only one!", searchResponse.getHits().totalHits,
						exerciseId);
			}
			if( searchResponse.getHits().getHits().length > 0) {
				return parseSearchHit(searchResponse.getHits().getHits()[0], objectMapper);
			}
		} catch (JsonProcessingException e) {
			throw new NotFoundException(" Exercise with id " + exerciseId + " not found?", e);
		} catch (IOException e) {
			throw new NotFoundException("Query for exercise with id " + exerciseId + " went wrong?", e);
		}
		throw new NotFoundException(" Exercise with id " + exerciseId + " not found?");
	}
}
